@@ -142,7 +142,7 @@ pub(crate) struct Vault {
     _vault_public_key: PublicKey,
     master_keys: HashMap<u32, EncryptionKey>,
     current_master_key_id: u32,
-    master_key_storage: Box<dyn AnyVaultKeyStorage>,
+    master_key_storage: Arc<dyn AnyVaultKeyStorage>,
 }
 
 impl Debug for Vault {
@@ -195,7 +195,7 @@ impl Vault {
     pub async fn initialize(
         server_id: StorageId,
         server_directory: &Path,
-        master_key_storage: Box<dyn AnyVaultKeyStorage>,
+        master_key_storage: Arc<dyn AnyVaultKeyStorage>,
     ) -> Result<Self, Error> {
         let master_keys_path = server_directory.join("master-keys");
         if master_keys_path.exists() {
@@ -209,7 +209,7 @@ impl Vault {
     async fn initialize_vault_key_storage(
         master_keys_path: &Path,
         server_id: StorageId,
-        master_key_storage: Box<dyn AnyVaultKeyStorage>,
+        master_key_storage: Arc<dyn AnyVaultKeyStorage>,
     ) -> Result<Self, Error> {
         let master_key = EncryptionKey::random();
         let (private, public) = DhP256HkdfSha256::gen_keypair(&mut thread_rng());
@@ -286,7 +286,7 @@ impl Vault {
     async fn unseal(
         master_keys_path: &Path,
         server_id: StorageId,
-        master_key_storage: Box<dyn AnyVaultKeyStorage>,
+        master_key_storage: Arc<dyn AnyVaultKeyStorage>,
     ) -> Result<Self, Error> {
         // The vault has been initilized previously. Do not overwrite this file voluntarily.
         let encrypted_master_keys = tokio::fs::read(master_keys_path)
@@ -704,7 +704,7 @@ mod tests {
             _vault_public_key: PublicKey::P256(public_key),
             master_keys,
             current_master_key_id: 0,
-            master_key_storage: Box::new(NullKeyStorage),
+            master_key_storage: Arc::new(NullKeyStorage),
         }
     }
 
@@ -739,21 +739,3 @@ mod tests {
         ));
     }
 }
-
-#[derive(Debug, Clone)]
-pub(crate) struct TreeVault {
-    pub key: KeyId,
-    pub vault: Arc<Vault>,
-}
-
-impl nebari::Vault for TreeVault {
-    type Error = crate::Error;
-
-    fn encrypt(&self, payload: &[u8]) -> Result<Vec<u8>, crate::Error> {
-        self.vault.encrypt_payload(&self.key, payload, None)
-    }
-
-    fn decrypt(&self, payload: &[u8]) -> Result<Vec<u8>, crate::Error> {
-        self.vault.decrypt_payload(payload, None)
-    }
-}
